
Imports System.Collections.Generic

Public Class SyntaxRichTextBox
    Inherits System.Windows.Forms.RichTextBox

    Friend TextModified As Boolean = False
    Friend Editing As Boolean
    Friend AutoSelect As Boolean = True
    Friend AutoComplete As Boolean = True

    ' ===============================================================================================
    '  DISABLE MOUSE SELECTION WHEN NOT EDITING
    ' ===============================================================================================
    Private Sub RTB_MouseLeave(ByVal sender As Object, ByVal e As System.EventArgs) Handles Me.MouseLeave
        If Not Editing Then Enabled = True
    End Sub
    Private Sub RTB_MouseDown(ByVal sender As Object, ByVal e As System.Windows.Forms.MouseEventArgs) Handles Me.MouseDown
        If Not Editing Then Enabled = False
    End Sub

    ' ===============================================================================================
    '  ENABLE DISABLE PAINT
    ' ===============================================================================================
    Friend PaintEventEnabled As Boolean = True
    Protected Overrides Sub WndProc(ByRef m As Message)
        Const WM_PAINT As Int32 = &HF
        If m.Msg = WM_PAINT AndAlso Not PaintEventEnabled Then
            m.Result = IntPtr.Zero
        Else
            MyBase.WndProc(m)
        End If
    End Sub



    ' ===============================================================================================================
    '   RTB - SAVE and RESTORE selection and scroll ( scroll aligned with the first text line )
    ' ===============================================================================================================
    Private RTB_OldSelectionStart As Int32
    Private RTB_OldSelectionLength As Int32
    Private RTB_OldFirstChar As Int32
    Private RTB_OldLastChar As Int32
    Private RTB_selection_saved As Int32 = 0
    Private Sub RTB_SaveSelectionAndScroll()
        If RTB_selection_saved <= 0 Then
            RTB_OldSelectionStart = Me.SelectionStart
            RTB_OldSelectionLength = Me.SelectionLength
            RTB_OldFirstChar = Me.GetCharIndexFromPosition(New Point(1, 1))
            RTB_OldLastChar = Me.GetCharIndexFromPosition(New Point(1, Me.ClientSize.Height))
            RTB_selection_saved = 1
        Else
            RTB_selection_saved += 1
        End If
    End Sub
    Private Sub RTB_RestoreSelectionAndScroll()
        RTB_selection_saved -= 1
        If RTB_selection_saved = 0 Then
            Me.Select(RTB_OldFirstChar, 0)
            Me.ScrollToCaret()
            If RTB_OldSelectionStart >= RTB_OldFirstChar And _
               RTB_OldSelectionStart <= RTB_OldLastChar Then
                Me.Select(RTB_OldSelectionStart, RTB_OldSelectionLength)
            End If
        End If
    End Sub


    ' ===============================================================================================
    '  LINES COLORS
    ' ===============================================================================================
    Friend Sub ColorAllLines()
        PaintEventEnabled = False
        Enabled = False
        ' ---------------------------------------------------------- save position 
        RTB_SaveSelectionAndScroll()
        ' ---------------------------------------------------------- 
        For i As Integer = 0 To Lines.Length - 1
            SelectLine(i)
            SelectionColor = Color_Text
            SelectionBackColor = Color_ActiveBack
            ' ------------------------------------------------------ AutomationLanguageColors
            AutomationLanguageColors(i, False)
        Next
        ' ---------------------------------------------------------- restore
        RTB_RestoreSelectionAndScroll()
        ' ---------------------------------------------------------- 
        Enabled = True
        PaintEventEnabled = True
        Me.Focus()
    End Sub

    Friend Sub ColorVisibleLines()
        ' ---------------------------------------------------------- color visible lines
        PaintEventEnabled = False
        Enabled = False
        ' ---------------------------------------------------------- save position 
        RTB_SaveSelectionAndScroll()
        ' ---------------------------------------------------------- 
        For i As Integer = GetFirstVisibleLine() To GetLastVisibleLine()
            SelectLine(i)
            SelectionColor = Color_Text
            SelectionBackColor = Color_ActiveBack
            ' ------------------------------------------------------ AutomationLanguageColors
            AutomationLanguageColors(i, True)
        Next
        ' ---------------------------------------------------------- restore
        RTB_RestoreSelectionAndScroll()
        ' ---------------------------------------------------------- 
        Enabled = True
        PaintEventEnabled = True
        Me.Focus()
    End Sub

    Friend Sub ColorAllOrVisibleLines()
        If Lines.Length < 200 Then
            ColorAllLines()
            Form1.CheckBox_ColorLines.Visible = False
        Else
            ColorVisibleLines()
            Form1.CheckBox_ColorLines.Visible = True
        End If
    End Sub

    Private Sub ColorLines(ByVal startline As Int32, ByVal endline As Int32)
        For i As Int32 = startline To endline
            ColorSingleLine(i)
        Next
    End Sub

    Private Sub ColorSingleLine(ByVal i As Int32)
        ' ---------------------------------------------------------- save position 
        Dim OldCursorPosition As Integer = SelectionStart
        ' ---------------------------------------------------------- set the whole line to default colors
        SelectLine(i)
        SelectionColor = Color_Text
        SelectionBackColor = Color_ActiveBack
        ' ---------------------------------------------------------- AutomationLanguageColors
        AutomationLanguageColors(i, False)
        ' ---------------------------------------------------------- restore
        SelectionStart = OldCursorPosition
        SelectionLength = 0
    End Sub

    Private LastEditedLine As Int32 = -1
    Protected Overrides Sub OnTextChanged(ByVal e As EventArgs)
        If Not PaintEventEnabled Then Return
        Dim line As Int32 = GetLineFromCharIndex(SelectionStart)
        ' ------------------------------------------------------- stop Paint
        PaintEventEnabled = False
        ' ------------------------------------------------------- process the line
        ColorSingleLine(line)
        ' ------------------------------------------------------- restart Paint
        PaintEventEnabled = True
        ' ------------------------------------------------------- 
        TextModified = True
        ' -------------------------------------------------------
        If line <> LastEditedLine Then
            Undo_AddEntry()
            LastEditedLine = line
        End If
    End Sub

    ' ===============================================================================================
    '  AUTOMATION LANGUAGE COLORS
    ' ===============================================================================================
    Private Color_Text As Color = Color.FromArgb(0, 0, 30)
    Private Color_ActiveBack As Color = Color.FromArgb(255, 255, 190)
    Private Color_CommentBack As Color = Color.FromArgb(240, 255, 240)
    Private Color_ErrorBack As Color = Color.FromArgb(255, 200, 180)
    Private Color_UnusedBack As Color = Color.FromArgb(60, 60, 80)

    Private Sub AutomationLanguageColors(ByVal i As Int32, ByVal fast As Boolean)
        If i < 0 Or i > Lines.Length - 1 Then Return
        ' --------------------------------------------------------------- 
        Dim line As String = Lines(i)
        ' --------------------------------------------------------------- comment position
        Dim CommentPosition As Int32 = RemoveComments(line)
        If CommentPosition = 0 Then
            MarkCommentLine(i, 0)
            Return
        End If
        ' --------------------------------------------------------------- 
        If line <> "" Then
            ' ----------------------------------------------------------- replace keywords
            'Dim sw As Stopwatch = New Stopwatch
            'sw.Start()
            If Not fast Then ReplaceKeywords(i, line)
            'Form1.Text = sw.Elapsed.TotalMilliseconds.ToString
            ' ----------------------------------------------------------- Tokenize to array of keywords
            Dim l() As String = Tokenize(line.Trim.ToLowerInvariant)
            ' ----------------------------------------------------------- 
            Select Case l.Length
                Case 0
                Case 1
                    Select Case l(0)
                        Case "beep", "else", "endif", "return", "stop"
                        Case Else : MarkErrorLine(i)
                    End Select
                Case 2
                    Select Case l(0)
                        Case "if", "wait"
                            If Not IsValidBoolean(l(1)) Then MarkErrorLine(i)
                        Case "label", "print", "gosub", "goto"
                        Case "load"
                            If l(1) <> "stop" And l(1) <> "hide" And Not l(1).StartsWith("http") Then
                                Select Case Strings.Right(l(1), 4).ToLower
                                    Case ".txt", ".jpg", ".bmp", ".gif", ".avi", ".wmv", ".mpg", ".mp4", ".mov", ".mp3", ".wav"
                                    Case Else : MarkErrorLine(i)
                                End Select
                            End If
                        Case "speed"
                            If Not IsValidDouble(l(1)) Then MarkErrorLine(i)
                        Case Else : MarkErrorLine(i)
                    End Select
                Case 3
                    Select Case l(0)
                        Case "button"
                            If Not IsNumeric(l(1)) Then MarkErrorLine(i)
                            Select Case l(2)
                                Case "enabled", "disabled"
                                Case Else : MarkErrorLine(i)
                            End Select
                        Case "wait"
                            Select Case l(1)
                                Case "button"
                                Case "seconds"
                                    If Not IsValidDouble(l(2)) Then MarkErrorLine(i)
                                Case Else : MarkErrorLine(i)
                            End Select
                        Case Else
                            If IsValidVarName(l(0)) AndAlso l(1) = "=" AndAlso IsValidDouble(l(2)) Then
                            ElseIf IsValidStringVarName(l(0)) AndAlso l(1) = "=" AndAlso IsValidString(l(2)) Then
                            Else
                                MarkErrorLine(i)
                            End If
                    End Select
                Case 4
                    Select Case l(0)
                        Case "button"
                            If Not IsNumeric(l(1)) Then MarkErrorLine(i)
                            Select Case l(2)
                                Case "text", "slot"
                                Case Else : MarkErrorLine(i)
                            End Select
                        Case "key"
                            If Not IsValidKey(l(1)) OrElse (l(2) <> "goto" AndAlso l(2) <> "gosub") Then
                                MarkErrorLine(i)
                            End If
                        Case "slot"
                            If IsValidDouble(l(1)) And l(2) = "=" And IsValidDouble(l(3)) Then
                            Else
                                Select Case l(3)
                                    Case "reset", "sleep"
                                    Case Else : MarkErrorLine(i)
                                End Select
                            End If
                        Case Else : MarkErrorLine(i)
                    End Select
                Case Else
                    MarkErrorLine(i)
            End Select
        End If
        ' ----------------------------------------------------------- color the comment section
        If CommentPosition > -1 Then
            MarkCommentLine(i, CommentPosition)
        End If
    End Sub
    Private Sub MarkErrorLine(ByVal i As Int32)
        SelectLine(i)
        SelectionBackColor = Color_ErrorBack
    End Sub
    Private Sub MarkCommentLine(ByVal i As Int32, Optional ByVal FirstCommentedChar As Int32 = -1)
        SelectLine(i)
        If FirstCommentedChar > 0 Then
            SelectionLength -= FirstCommentedChar ' lenght must be changed before start
            SelectionStart += FirstCommentedChar  ' otherwise at the end of the file the lenght is shortened 
        End If
        SelectionBackColor = Color_CommentBack
    End Sub

    ' ===============================================================================================
    '  MIXED FUNCTIONS
    ' ===============================================================================================
    Friend Function IsValidKey(ByVal KeyString As String) As Boolean
        Select Case KeyString
            Case "left", "right", "up", "down", "space", "esc"
                Return True
            Case Else
                If KeyString.Length = 1 Then
                    If InStr("0123456789abcdefghijklmnopqrstuvwxyz", KeyString) >= 0 Then Return True
                End If
        End Select
        Return False
    End Function
    Friend Function IsValidVarName(ByVal VarName As String) As Boolean
        If VarName.Length <> 2 Then Return False
        If Not VarName.StartsWith("v", StringComparison.InvariantCultureIgnoreCase) Then Return False
        If Not "123456789".Contains(VarName(1)) Then Return False
        Return True
    End Function
    Friend Function IsValidStringVarName(ByVal VarName As String) As Boolean
        If VarName.Length <> 2 Then Return False
        If Not VarName.StartsWith("s", StringComparison.InvariantCultureIgnoreCase) Then Return False
        If Not "123456789".Contains(VarName(1)) Then Return False
        Return True
    End Function
    Friend Function VarNameToIndex(ByVal VarName As String) As Int32
        If VarName.Length <> 2 Then Return 0
        Return CInt(Val(VarName(1)))
    End Function
    Friend Function ExtractFirstWord(ByVal s As String) As String
        If s = Nothing Then Return ""
        s = s.Replace(vbTab, "").Trim
        Dim i As Int32 = s.IndexOf(" ")
        If i < 0 Then Return s
        Return s.Substring(0, i)
    End Function
    Friend Function RemoveComments(ByRef s As String) As Int32
        Dim i As Int32 = s.IndexOf("'")
        If i > 0 Then s = s.Remove(i).TrimEnd
        If i = 0 Then s = ""
        Return i
    End Function
    Friend Sub SelectLine(ByVal line As Int32)
        If line >= 0 AndAlso line < Lines.Length AndAlso Lines.Length > 0 Then
            Dim lineStart As Int32 = GetFirstCharIndexFromLine(line)
            Dim lineLength As Int32 = Lines(line).Length
            Me.Select(lineStart, lineLength)
        Else
            Me.Select(0, 0)
        End If
    End Sub
    Private Sub SelectLines(ByVal FirstLine As Int32, ByVal LastLine As Int32)
        Dim FirstChar As Int32 = Me.GetFirstCharIndexFromLine(FirstLine)
        Dim LastChar As Int32 = Me.GetFirstCharIndexFromLine(LastLine) + Me.Lines(LastLine).Length '- 1
        Me.Select(FirstChar, LastChar - FirstChar)
    End Sub
    Private Function GetFirstSelectedLine() As Int32
        Return Me.GetLineFromCharIndex(Me.SelectionStart)
    End Function
    Private Function GetLastSelectedLine() As Int32
        Return Me.GetLineFromCharIndex(Me.SelectionStart + Me.SelectionLength)
    End Function
    Friend Function GetFirstVisibleLine() As Int32
        Return Me.GetLineFromCharIndex(Me.GetCharIndexFromPosition(New Point(1, 1)))
    End Function
    Friend Function GetLastVisibleLine() As Int32
        Dim p As Point = New Point(1, Me.ClientSize.Height)
        Return Me.GetLineFromCharIndex(Me.GetCharIndexFromPosition(p))
    End Function
    Private Function TestEndif(ByVal lnum As Int32) As Boolean
        Dim counter As Int32 = 0
        Dim line As String
        For i As Int32 = lnum + 1 To Me.Lines.Length - 1
            line = Me.Lines(i).Trim.ToLower
            If line = "endif" Then
                If counter <= 0 Then
                    Return True
                End If
                counter -= 1
            ElseIf line.StartsWith("if") Then
                counter += 1
            End If
        Next
        Return False
    End Function
    Private Function TestElseOrEndif(ByVal lnum As Int32) As Boolean
        Dim counter As Int32 = 0
        Dim line As String
        For i As Int32 = lnum + 1 To Me.Lines.Length - 1
            line = Me.Lines(i).Trim.ToLower
            If line = "else" Then
                If counter <= 0 Then
                    Return True
                End If
            ElseIf line = "endif" Then
                If counter <= 0 Then
                    Return True
                End If
                counter -= 1
            ElseIf line.StartsWith("if") Then
                counter += 1
            End If
        Next
        Return False
    End Function

    ' ===============================================================================================
    '  TOKENIZER
    ' ===============================================================================================
    Enum TokenizerStatus As Int32
        Normal = 0
        InsideToken = 1
        InsideString = 2
    End Enum
    Friend Function Tokenize(ByVal line As String) As String()
        Dim status As TokenizerStatus
        Dim l(-1) As String
        For i As Int32 = 0 To line.Length - 1
            Select Case status
                Case TokenizerStatus.Normal
                    Select Case line(i)
                        Case " "c, Chr(9), "("c, ")"c
                        Case "'"c
                            Exit For ' comments are completely skipped 
                        Case """"c
                            status = TokenizerStatus.InsideString
                            ReDim Preserve l(l.Length)
                        Case Else
                            status = TokenizerStatus.InsideToken
                            ReDim Preserve l(l.Length)
                            l(l.Length - 1) += line(i)
                    End Select
                Case TokenizerStatus.InsideToken
                    Select Case line(i)
                        Case " "c, Chr(9), "("c, ")"c
                            ' ------------------------------------------ end of token - eval right side ?  
                            status = TokenizerStatus.Normal
                            Select Case l.Length
                                Case 1
                                    Select Case l(0).ToLower
                                        Case "if", "print", "speed"
                                            ' -------------------------- isolate right side formula and exit
                                            ReDim Preserve l(1)
                                            l(1) = line.Substring(i + 1)
                                            RemoveComments(l(1))
                                            Exit For
                                        Case "slot"
                                            ' -------------------------- isolate left side formula and continue
                                            Dim eq As Int32 = InStr(line, "=")
                                            If eq > i Then
                                                ReDim Preserve l(1)
                                                l(1) = line.Substring(i, eq - i - 2)
                                                i = eq - 2
                                            End If
                                    End Select
                                Case 2
                                    Select Case l(0).ToLower
                                        Case "wait"
                                            If l(1).ToLower <> "seconds" And l(1).ToLower <> "button" Then
                                                ' -------------------------- isolate right side formula and exit
                                                ReDim Preserve l(1)
                                                l(1) = line.Substring(5)
                                                RemoveComments(l(1))
                                                Exit For
                                            End If
                                        Case "v1", "v2", "v3", "v4", "v5", "v6", "v7", "v8", "v9", _
                                             "s1", "s2", "s3", "s4", "s5", "s6", "s7", "s8", "s9"
                                            If l(1) = "=" Then
                                                ' -------------------------- isolate right side formula and exit
                                                ReDim Preserve l(2)
                                                l(2) = line.Substring(5)
                                                RemoveComments(l(2))
                                                Exit For
                                            End If
                                    End Select
                                Case 3
                                    Select Case l(0).ToLower
                                        Case "slot"
                                            If l(2) = "=" Then
                                                ' ----------------------- isolate right side formula and exit
                                                ReDim Preserve l(3)
                                                l(3) = line.Substring(i + 1)
                                                RemoveComments(l(3))
                                                Exit For
                                            End If
                                        Case "wait"
                                            If l(1).ToLower = "seconds" Then
                                                ' -------------------------- isolate right side formula and exit
                                                ReDim Preserve l(2)
                                                l(2) = line.Substring(13)
                                                RemoveComments(l(2))
                                                Exit For
                                            End If

                                    End Select
                            End Select
                        Case Else
                            l(l.Length - 1) += line(i)
                    End Select
                Case TokenizerStatus.InsideString
                    Select Case line(i)
                        Case """"c
                            If i >= line.Length - 2 OrElse line(i + 1) <> """"c Then
                                status = TokenizerStatus.Normal
                            Else
                                l(l.Length - 1) += line(i)
                                i += 1
                            End If
                        Case Else
                            l(l.Length - 1) += line(i)
                    End Select
            End Select
        Next
        Return l
    End Function


    ' ===============================================================================================
    '  REPLACE KEYWORDS
    ' ===============================================================================================
    ' Short keywords are listed before
    Private Keywords() As String = {"PI", "Int", _
                                    "Beep", "Button", "Disabled", "Enabled", "GoSub", "GoTo", _
                                    "If", "Else", "EndIf", "Key", "Label", "Load", "Print", "Return", _
                                    "Seconds", "Sleep", "Slot", "Speed", _
                                    "Stop", "Text", "Wait", "Hide", _
                                    "Left", "Right", "Up", "Down", "Space", "Esc", _
                                    "Sin", "Cos", "Asin", "Acos", "Tan", "Atan", "Atan2", "Pow", _
                                    "Round", "Mid", "Len", _
                                    "Trim", "Format", "UCase", "LCase", "WCase", _
                                    "Date", "Now", "Today", "Rnd", "Reset", "Sleep"}
    Private Sub ReplaceKeywords(ByVal lineNumber As Int32, ByVal line As String)
        Dim position As Int32
        Dim LineStart As Int32 = GetFirstCharIndexFromLine(lineNumber)
        For Each kw As String In Keywords
            position = 0
            Do
                position = line.IndexOf(kw, position, StringComparison.InvariantCultureIgnoreCase)
                If position >= 0 Then
                    If Not InsideQuotes(position, line) Then
                        Me.Select(LineStart + position, kw.Length)
                        If kw <> Me.SelectedText Then
                            Me.SelectedText = kw
                        End If
                    End If
                    position += kw.Length
                End If
            Loop Until position < 0
        Next
    End Sub
    Private Function InsideQuotes(ByVal index As Int32, ByVal text As String) As Boolean
        InsideQuotes = False
        If index >= text.Length Then Return False
        For Counter As Int32 = 0 To index - 1
            If text(Counter) = """"c Then
                InsideQuotes = Not InsideQuotes
            End If
        Next
    End Function

    Friend Sub ShowLinesCount()
        Dim nline As Int32 = GetLineFromCharIndex(SelectionStart)
        Form1.Label_Lines.Text = "Lines: " + Me.Lines.Length.ToString + "    Line: " + (nline + 1).ToString
    End Sub

    ' ===============================================================================================
    '  SUGGESTIONS
    ' ===============================================================================================
    Private Sub RTB_SelectionChanged(ByVal sender As Object, ByVal e As System.EventArgs) Handles Me.SelectionChanged
        If Not EventsAreEnabled Then Exit Sub
        ShowSuggestions(Form1.StatusLabel1, Form1.Label_Lines)
    End Sub
    Friend Sub ShowSuggestions(ByVal label1 As Label, ByVal label2 As Label)
        If Not Editing Then Return
        Dim nline As Int32 = GetLineFromCharIndex(SelectionStart)
        ShowLinesCount()
        If Lines.Length = 0 Then Return
        ' ----------------------------------------------------------- trimmed amd lower case
        Dim line As String = Lines(nline).Trim.ToLower
        ' ----------------------------------------------------------- Tokenize to array of keywords 
        RemoveComments(line)
        Dim l() As String = Tokenize(line)
        Dim s As String = ""
        Select Case l.Length
            Case 0
                s = "Beep  Button  GoSub  GoTo  If  Else  EndIf  Key  Label  Load  Print  Return  Slot  Speed  Stop  Wait"
            Case 1
                Select Case l(0)
                    Case "beep" : s = "OK: A system Beep will be emitted here (repetition is limited to 10 per second)"
                    Case "button" : s = "BUTTON-NUMBER"
                    Case "gosub" : s = "LABEL-NAME (use "" to include spaces in the text)"
                    Case "goto" : s = "LABEL-NAME (use "" to include spaces in the text)"
                    Case "if" : s = "V1 to V9"
                    Case "else"
                        If TestEndif(nline) Then
                            s = "OK: Following instructions are executed when the preceding IF is FALSE"
                        Else
                            s = "WARNING: This line has no effect without a corresponding EndIf"
                        End If
                    Case "endif" : s = "OK: This is the end of the conditional execution."
                    Case "key" : s = "0..9 / a..z / F1..F12 / Left / Right / Up / Down / Space / Esc"
                    Case "label" : s = "LABEL-NAME (use "" to include spaces in the text)"
                    Case "load" : s = "Stop Hide or FILENAME(.txt, .jpg, .bmp, .gif, .avi, .wmv, .mpg, .mp4, .mov, .mp3, .wav) or HTTP"
                    Case "print" : s = "OK: ""TEXT TO PRINT"" (also complex expressions - use "" to include spaces in the text)"
                    Case "return" : s = "OK: The program will return to the line after ""GoSub"" or ""Button"" events"
                    Case "slot" : s = "EXPRESSION"
                    Case "speed" : s = "1 to 9 - Normal values are: 5(slow execution) 7(fast execution) 9(only for short loops)"
                    Case "stop" : s = "OK: The program will stop here"
                    Case "wait" : s = "Seconds / Button / Condition"
                    Case Else
                        If IsValidVarName(l(0)) Then
                            s = "="
                        Else
                            s = "Beep  Button  GoSub  GoTo  If  Else  EndIf  Key  Label  Load  Print  Return  Slot  Speed  Stop  Wait"
                        End If
                End Select
            Case 2
                Select Case l(0)
                    Case "button"
                        If IsNumeric(l(1)) Then
                            s = "Slot / Text / Enabled / Disabled"
                        Else
                            s = "ERROR: Invalid button number"
                        End If
                    Case "gosub"
                        s = "OK: The program will call the subprogram starting with the label """ + l(1) + """ (if exists)"
                    Case "goto"
                        s = "OK: The program will jump to the label """ + l(1) + """ (if exists)"
                    Case "if"
                        If IsValidBoolean(l(1)) Then
                            If TestElseOrEndif(nline) Then
                                s = "OK: the expression """ + l(1) + """ will be evalued"
                            Else
                                s = "WARNING: This line has no effect without a corresponding Else or EndIf"
                            End If
                        Else
                            s = "ERROR: " + LastError
                        End If
                    Case "key"
                        If IsValidKey(l(1)) Then
                            s = "Goto / Gosub"
                        Else
                            s = "Please use a valid Key: 0..9 / a..z / F1..F12 / Left / Right / Up / Down / Space / Esc"
                        End If
                    Case "label"
                        s = "OK: This label will be called by ""GoSub"", ""GoTo"" and by ""Buttons"" with text """ + l(1) + """"
                    Case "load"
                        If l(1) = "stop" Then
                            s = "OK: Sounds and video will be stopped here"
                        ElseIf l(1) = "hide" Then
                            s = "OK: The running program will be visualized until a new image or video will be loaded"
                        Else
                            If l(1) IsNot Nothing AndAlso l(1).Length > 3 Then
                                Select Case Strings.Right(l(1), 4).ToLower
                                    Case ".txt" : s = "OK: The program will be loaded from ""Programs"" folder"
                                    Case ".jpg", ".bmp", ".gif" : s = "OK: The image will be loaded from ""Media"" folder"
                                    Case ".avi", ".wmv", ".mpg", ".mp4", ".mov" : s = "OK: The video will be loaded from ""Media"" folder"
                                    Case ".mp3", ".wav" : s = "OK: The audio file will be played (with the video player)"
                                    Case Else : s = "Stop Hide or FILENAME(.txt, .jpg, .bmp, .gif, .avi, .wmv, .mpg, .mp4, .mov, .mp3, .wav) or HTTP"
                                End Select
                            Else
                                s = "Stop Hide or FILENAME(.txt, .jpg, .bmp, .gif, .avi, .wmv, .mpg, .mp4, .mov, .mp3, .wav) or HTTP"
                            End If
                        End If
                    Case "print"
                        s = "OK: The expression """ + l(1) + """ will be evalued and printed"
                    Case "slot"
                        If IsValidDouble(l(1)) Then
                            s = "Please write ""="""
                        Else
                            s = "Please write a valid numerical expression"
                        End If
                    Case "speed"
                        If IsValidDouble(l(1)) Then
                            s = "OK: The speed """ + l(1) + """ will be used"
                        Else
                            s = "Please write 1 to 9 or a valid numeric expression"
                        End If
                    Case "wait"
                        Select Case l(1)
                            Case "seconds" : s = "SECONDS (also with decimals)"
                            Case "button" : s = "BUTTON-NAME"
                            Case Else
                                If IsValidBoolean(l(1)) Then
                                    s = "OK: This line waits until the expression """ + l(1) + """ will be TRUE"
                                Else
                                    s = "Please write a formula with a condition - TRUE to continue"
                                End If
                        End Select
                    Case Else
                        If IsValidVarName(l(0)) AndAlso l(1) = "=" Then
                            s = "Please write a valid numerical expression"
                        Else
                            s = "ERROR"
                        End If
                End Select
            Case 3
                Select Case l(0)
                    Case "button"
                        If IsNumeric(l(1)) Then
                            Select Case l(2)
                                Case "enabled", "disabled"
                                    s = "OK: Button will be " + l(2)
                                Case "slot" : s = "SLOT-NUMBER"
                                Case "text" : s = "BUTTON-NAME (use "" to include spaces in the name)"
                                Case Else : s = "Slot / Text / Enabled / Disabled"
                            End Select
                        Else
                            s = "ERROR: Invalid button number"
                        End If
                    Case "key"
                        If IsValidKey(l(1)) Then
                            If l(2) = "goto" OrElse l(2) = "gosub" Then
                                s = "Please write a Label name"
                            Else
                                s = "Goto / Gosub"
                            End If
                        Else
                            s = "ERROR: The key is not valid"
                        End If
                    Case "slot"
                        If IsValidDouble(l(1)) Then
                            If l(2) = "=" Then
                                s = "Please write RESET, SLEEP or a valid numerical expression"
                            Else
                                s = "Please write a ""="""
                            End If
                        Else
                            s = "The slot number is not valid"
                        End If
                    Case "wait"
                        Select Case l(1)
                            Case "seconds"
                                If IsValidDouble(l(2)) Then
                                    s = "OK: The program will be paused for """ + l(2) + """ seconds"
                                Else
                                    s = "Please write SECONDS (also with decimals) or a valid expression"
                                End If
                            Case "button"
                                s = "OK: The program will wait the button """ + l(2) + """ (for buttons with long names the first word is used)"
                            Case Else
                                s = "OK: The expression """ + l(2) + """ will be evaluated."
                        End Select
                    Case Else
                        If IsValidVarName(l(0)) AndAlso l(1) = "=" AndAlso IsValidDouble(l(2)) Then
                            s = "OK: Numeric variable """ + l(0) + """ = " + l(2)
                        ElseIf IsValidStringVarName(l(0)) AndAlso l(1) = "=" AndAlso IsValidString(l(2)) Then
                            s = "OK: String variable """ + l(0) + """ = " + l(2)
                        Else
                            s = "ERROR"
                        End If
                End Select
            Case 4
                Select Case l(0)
                    Case "button"
                        If Not IsNumeric(l(1)) Then
                            s = "ERROR"
                        Else
                            Select Case l(2)
                                Case "slot"
                                    s = "OK: Button """ + l(1) + """ will be pressed by slot """ + l(3) + """ (when value exceeds 500)"
                                Case "text"
                                    s = "OK: Button """ + l(1) + """ text is """ + l(3) + """ (the first word """ + ExtractFirstWord(l(3)) + """ will be used for ""Labels"" and ""Wait Button"")"
                                Case Else : s = "ERROR"
                            End Select
                        End If
                    Case "key"
                        If IsValidKey(l(1)) Then
                            If l(2) = "goto" Then
                                s = "OK: Pressing the key """ + l(1) + """ the Label """ + l(3) + """ will be called by a GoTo"
                            ElseIf l(2) = "gosub" Then
                                s = "OK: Pressing the key """ + l(1) + """ the Label """ + l(3) + """ will be called by a GoSub (remember to use a Return)"
                            Else
                                s = "Goto / Gosub"
                            End If
                        Else
                            s = "ERROR: The key is not valid"
                        End If
                    Case "slot"
                        If IsValidDouble(l(1)) And l(2) = "=" And IsValidDouble(l(3)) Then
                            s = "OK: Slot """ + l(1).Trim + """ = expression: """ + l(3).Trim + """"
                        Else
                            Select Case l(3)
                                Case "reset", "sleep"
                                    s = "OK: Slot """ + l(1).Trim + """ will be set with the special value: NAN+" + l(3)
                                Case Else
                                    s = "Please write RESET , SLEEP or a valid expression"
                            End Select
                        End If
                End Select
            Case Else
                s = "ERROR"
        End Select
        ' -----------------------------------------------------------
        If s.StartsWith("OK") Then
            label1.BackColor = Color.FromArgb(250, 255, 200)
            label1.Text = s
        ElseIf s.StartsWith("ERROR") Then
            label1.BackColor = Color.FromArgb(255, 160, 150)
            label1.Text = s
        ElseIf s.StartsWith("WARNING") Then
            label1.BackColor = Color.FromArgb(255, 200, 100)
            label1.Text = s
        ElseIf s = "" Then
            label1.BackColor = Color.AliceBlue
            label1.Text = ""
        Else
            If AutoSelect AndAlso l.Length > 0 Then
                ' ------------------------------------------------------ Keyword auto selection
                Dim s2 As String = ExtractKeywords(s, GetCompleteWord())
                If s2 <> "" Then
                    s = s2
                    ' ------------------------------------------------------ Keyword auto completion
                    If AutoComplete Then
                        If Array.IndexOf(Keywords, s) >= 0 Then
                            Select Case LastKeyPressed
                                Case Keys.Back, Keys.Cancel, Keys.Clear, Keys.Delete, Keys.OemClear
                                Case Else
                                    CompleteKeyword(s)
                            End Select
                        End If
                    End If
                End If
            End If
            ' ------------------------------------------------------ Keyword suggestion
            label1.BackColor = Color.FromArgb(220, 255, 200)
            label1.Text = "Keywords: " & s
        End If
    End Sub


    ' ===============================================================================================
    '  AUTOMATIC COMPLETION
    ' ===============================================================================================
    Private Function ExtractKeywords(ByVal s1 As String, ByVal s2 As String) As String
        If s2 = "" Then Return s1
        ExtractKeywords = s1
        Dim MatchesCount As Int32 = 0
        Dim s() As String = Split(s1.Replace(" / ", " "), " ")
        s1 = ""
        For Each word As String In s
            If word.StartsWith(s2, StringComparison.InvariantCultureIgnoreCase) Then
                If s1 <> "" Then s1 += " / "
                s1 += word
            End If
        Next
        Return s1
    End Function
    Private Function GetCompleteWord() As String
        Dim start, length As Int32
        FindWordPointers(start, length)
        Return Me.Text.Substring(start, length)
    End Function
    Private Sub FindWordPointers(ByRef start As Int32, ByRef lenght As Int32)
        Const separators As String = " " + vbCr + vbLf + vbTab
        Dim i1, i2 As Int32
        For i1 = SelectionStart To 1 Step -1
            If InStr(separators, Me.Text(i1 - 1)) > 0 Then
                Exit For
            End If
        Next
        For i2 = SelectionStart To Me.Text.Length - 1
            If InStr(separators, Me.Text(i2)) > 0 Then
                Exit For
            End If
        Next
        start = i1
        lenght = i2 - i1
    End Sub
    Private Sub CompleteKeyword(ByVal keyword As String)
        ' --------------------------------------------------------------- save
        EventsAreEnabled = False
        PaintEventEnabled = False
        Dim OldCursorPosition As Integer = SelectionStart
        ' --------------------------------------------------------------- change
        Dim start, length As Int32
        FindWordPointers(start, length)
        SelectionStart = start
        SelectionLength = length
        SelectedText = keyword & " "
        ' --------------------------------------------------------------- restore
        SelectionLength = 0
        EventsAreEnabled = True
        PaintEventEnabled = True
        ' --------------------------------------------------------------- update suggestions
        SelectionStart = SelectionStart - 1
        SelectionStart = SelectionStart + 1
    End Sub

    ' ===============================================================================================
    '  CUT COPY and PASTE
    ' ===============================================================================================
    Friend Sub ExecCut()
        If Me.SelectionLength > 0 Then
            SendTextToClipboard()
            If UndoTextReady Then Undo_AddEntry()
            Me.SelectedText = ""
            'Undo_AddEntry()
        End If
    End Sub
    Friend Sub ExecCopy()
        If Me.SelectionLength > 0 Then
            SendTextToClipboard()
        End If
    End Sub
    Private Sub SendTextToClipboard()
        Dim data As IDataObject
        data = New DataObject(DataFormats.Text, Me.SelectedText.Replace(vbLf, vbCrLf))
        Clipboard.SetDataObject(data, True)
    End Sub
    Friend Sub ExecPaste()
        If Clipboard.GetDataObject().GetDataPresent(DataFormats.Text) Then
            If UndoTextReady Then Undo_AddEntry()
            Me.Enabled = False
            PaintEventEnabled = False
            ' ------------------------------------------------------------ extract only the text from clipbbard
            Dim startline As Int32 = GetLineFromCharIndex(SelectionStart)
            Me.SelectedText = Clipboard.GetText
            Dim endline As Int32 = GetLineFromCharIndex(SelectionStart + SelectionLength)
            ' ------------------------------------------------------------ color lines
            ColorLines(startline, endline)
            ' ------------------------------------------------------------
            PaintEventEnabled = True
            Me.Enabled = True
            Me.Focus()
            Undo_AddEntry()
        End If
    End Sub
    Friend Sub ExecDelete()
        If Me.SelectionLength > 0 Then
            If UndoTextReady Then Undo_AddEntry()
            Me.SelectedText = ""
            'Undo_AddEntry()
        End If
    End Sub
    'Friend Function Editor_CanPaste() As Boolean
    '    Return Clipboard.GetDataObject().GetDataPresent(DataFormats.Text)
    'End Function

    ' ===============================================================================================
    '  INDENT-UNINDENT
    ' ===============================================================================================
    Friend Sub IndentSelectedLines(ByVal increment As Int32)
        If UndoTextReady Then Undo_AddEntry()
        Me.Enabled = False
        PaintEventEnabled = False
        ' ----------------------------------------------------------------
        Dim FirstSelectedLine As Int32 = GetFirstSelectedLine()
        Dim LastSelectedLine As Int32 = GetLastSelectedLine()
        ' ----------------------------------------------------------------
        For i As Int32 = FirstSelectedLine To LastSelectedLine
            Dim lineText As String = Me.Lines(i)
            Dim ntabs As Int32 = CountInitialTabs(lineText, 4) + increment
            If ntabs >= 0 Then
                Dim start As Int32 = Me.GetFirstCharIndexFromLine(i)
                Me.Select(start, CountInitialChars(lineText))
                Me.SelectedText = StrDup(ntabs, vbTab)
            End If
        Next
        ' ----------------------------------------------------------------
        SelectLines(FirstSelectedLine, LastSelectedLine)
        ' ----------------------------------------------------------------
        Me.Enabled = True
        PaintEventEnabled = True
        Me.Focus()
        Undo_AddEntry()
    End Sub

    ' ===============================================================================================
    '  COMMENT UN-COMMENT
    ' ===============================================================================================
    Friend Sub CommentSelectedLines(ByVal comment As Boolean)
        If UndoTextReady Then Undo_AddEntry()
        Me.Enabled = False
        PaintEventEnabled = False
        ' ----------------------------------------------------------------
        Dim CommentColumn As Int32 = -1
        ' ----------------------------------------------------------------
        Dim FirstSelectedLine As Int32 = GetFirstSelectedLine()
        Dim LastSelectedLine As Int32 = GetLastSelectedLine()
        ' ----------------------------------------------------------------
        For i As Int32 = FirstSelectedLine To LastSelectedLine
            Dim lineText As String = Me.Lines(i)
            Dim firstch As Int32 = CountInitialChars(lineText)
            If firstch >= 0 Then
                If CommentColumn < 0 Then CommentColumn = firstch
                Dim start As Int32 = Me.GetFirstCharIndexFromLine(i)

                If comment Then
                    Me.Select(start + CommentColumn, 0)
                    Me.SelectedText = "'"
                Else
                    Me.Select(start + firstch, 1)
                    If Me.SelectedText = "'" Then
                        Me.SelectedText = ""
                    End If
                End If
            End If
        Next
        ' ----------------------------------------------------------------
        ColorLines(FirstSelectedLine, LastSelectedLine)
        ' ----------------------------------------------------------------
        Me.Enabled = True
        PaintEventEnabled = True
        Me.Focus()
        Undo_AddEntry()
    End Sub

    Private Function CountInitialChars(ByVal s As String) As Int32
        Dim CharCount As Int32 = 0
        For i As Int32 = 0 To s.Length - 1
            Select Case s.Substring(i, 1)
                Case vbTab : CharCount += 1
                Case " " : CharCount += 1
                Case Else : Exit For
            End Select
        Next
        Return CharCount
    End Function
    Private Function CountInitialTabs(ByVal s As String, ByVal TabSize As Int32) As Int32
        Dim SpaceCount As Int32 = 0
        Dim TabCount As Int32 = 0
        For i As Int32 = 0 To s.Length - 1
            Select Case s.Substring(i, 1)
                Case vbTab : TabCount += 1
                Case " " : SpaceCount += 1
                Case Else : Exit For
            End Select
        Next
        Return TabCount + SpaceCount \ TabSize
    End Function


    ' ===============================================================================================
    '  UNDO - Undo entries are saved only after important text changes
    ' ===============================================================================================
    Private LastKeyPressed As Keys
    Private UndoTextReady As Boolean = False
    Private Sub SyntaxRichTextBox_KeyDown(ByVal sender As Object, ByVal e As System.Windows.Forms.KeyEventArgs) Handles Me.KeyDown
        Select Case e.KeyCode
            Case Keys.Enter, Keys.Cancel, Keys.Delete
                If UndoTextReady Then Undo_AddEntry()
            Case Keys.Tab
                If SelectedText.Length > 0 Then
                    IndentSelectedLines(If(e.Shift, -1, 1))
                    e.Handled = True
                    e.SuppressKeyPress = True
                End If
        End Select
        LastKeyPressed = e.KeyCode
    End Sub
    Private Sub SyntaxRichTextBox_KeyUp(ByVal sender As Object, ByVal e As System.Windows.Forms.KeyEventArgs) Handles Me.KeyUp
        Select Case e.KeyCode
            Case Keys.Enter, Keys.Cancel, Keys.Delete
                'Undo_AddEntry()
            Case Keys.Left, Keys.Right, Keys.Up, Keys.Down
            Case Else
                UndoTextReady = True
        End Select
    End Sub
    Private Sub SyntaxRichTextBox_LostFocus(ByVal sender As Object, ByVal e As System.EventArgs) Handles Me.LostFocus
        If UndoTextReady Then
            Undo_AddEntry()
        End If
    End Sub

    ' ------------------------------------------------------------------- private undo functions
    Private Structure UndoEntry
        Dim ActionName As String
        Dim RTF_Text As String
        Dim FirstVisibleChar As Int32
        Dim SelectionStart As Int32
        'Dim SelectionLength As Int32
    End Structure
    Private UndoBuffer As List(Of UndoEntry) = New List(Of UndoEntry)
    Private Undo_Index As Int32
    Friend Sub Undo_Clear()
        UndoBuffer.Clear()
        Undo_Index = 0
        Undo_AddEntry()
        TextModified = False
    End Sub
    Friend Sub Undo_AddEntry()
        If Not PaintEventEnabled Then Exit Sub
        ' ----------------------------------------------------------------
        Dim newentry As UndoEntry
        newentry.FirstVisibleChar = Me.GetCharIndexFromPosition(New Point(1, 1))
        newentry.SelectionStart = Me.SelectionStart
        'newentry.SelectionLength = Me.SelectionLength
        newentry.ActionName = ""
        newentry.RTF_Text = Me.Rtf
        ' ----------------------------------------------------------------
        While UndoBuffer.Count > Undo_Index
            UndoBuffer.RemoveAt(UndoBuffer.Count - 1)
        End While
        ' ----------------------------------------------------------------
        UndoBuffer.Add(newentry)
        Undo_Index += 1
        'Debug.Print(UndoBuffer.Count.ToString)
        ' ----------------------------------------------------------------
        UndoTextReady = False
        TextModified = True
    End Sub

    ' ===============================================================================================
    '  UNDO and REDO commands
    ' ===============================================================================================
    Friend Sub ExecUndo()
        ' ---------------------------------------------------------------- 
        Me.Enabled = False
        PaintEventEnabled = False
        ' ---------------------------------------------------------------- undo the RTF text
        Undo_Index -= 1
        If Undo_Index < 1 Then Undo_Index = 1
        Dim undo As UndoEntry
        undo = UndoBuffer(Undo_Index - 1)
        Me.Rtf = undo.RTF_Text
        ' ---------------------------------------------------------------- update scroll and selection
        Me.Select(undo.FirstVisibleChar, 0)
        Me.ScrollToCaret()
        Me.Select(undo.SelectionStart, 0) 'undo.SelectionLength)
        ' ----------------------------------------------------------------
        Form1.ApplyEditorProps(True)
        Me.Enabled = True
        PaintEventEnabled = True
        ' ----------------------------------------------------------------  mark as not-modified
        If Undo_Index < 2 Then TextModified = False
    End Sub
    Friend Sub ExecRedo()
        ' ---------------------------------------------------------------- 
        Me.Enabled = False
        PaintEventEnabled = False
        ' ---------------------------------------------------------------- undo the RTF text
        If Undo_Index > UndoBuffer.Count - 1 Then Undo_Index = UndoBuffer.Count - 1
        Dim undo As UndoEntry = UndoBuffer(Undo_Index)
        Me.Rtf = undo.RTF_Text
        Undo_Index += 1
        ' ---------------------------------------------------------------- update scroll and selection
        Me.Select(undo.FirstVisibleChar, 0)
        Me.ScrollToCaret()
        Me.Select(undo.SelectionStart, 0)
        ' ---------------------------------------------------------------- 
        Form1.ApplyEditorProps(True)
        Me.Enabled = True
        PaintEventEnabled = True
        ' ---------------------------------------------------------------- mark as modified
        TextModified = True
    End Sub
    'Friend Function Editor_CanUndo() As Boolean
    '    Return Undo_Index >= 2
    'End Function
    'Friend Function Editor_CanRedo() As Boolean
    '    Return Undo_Index <= UndoBuffer.Count - 1
    'End Function

End Class
